#include "qkd/copy.h"
#include "qkd/interfaces.h"
#include "qkd/models.h"
#include "qkd/ssl_socket_server.h"
#include "qkd/sizes.h"


typedef struct qkd_ksid_info {
    uuid_t key_stream_id;
    qkd_qos_t qos;
    qkd_uri_t source;
    qkd_uri_t destination;
    struct qkd_ksid_info *ksid_info_next;
} qkd_ksid_info_t;

typedef struct server_context_data_example {
    int key_rate;
    float qber;
    char *example_db_name;
    uuid_t node_uuid;
    qkd_ksid_info_t *ksid_info_linked_list;
} server_context_data_example_t;

qkd_key_buffer_t gen_pseudorandom_key(qkd_uri_t source, qkd_uri_t destination, size_t key_size, uint32_t index) {
    qkd_key_buffer_t key = {key_size, malloc(key_size)};
    memset(key.data, 0x22, key_size);
    return key;
}

void qkd_open_connect(qkd_uri_t source, qkd_uri_t destination, qkd_qos_t *qos, uuid_t *key_stream_id,
                      qkd_status_t *status, void *app_context_data) {
    if (uuid_compare(*key_stream_id, (uuid_t) {}) == 0)
        uuid_generate_random(*key_stream_id);  // Generate random if KSID is 00000000-0000-0000-0000-000000000000
    *qos = *qos;

    // Use or modify app_context_data
    server_context_data_example_t *config = app_context_data;
    qkd_ksid_info_t *ksid_info_list = config->ksid_info_linked_list;

    while (ksid_info_list != NULL) {  // Check KSID does not exist
        if (uuid_compare(*key_stream_id, ksid_info_list->key_stream_id) == 0) {
            *status = qkd_status_code_oc_ksid_already_in_use;
            return;
        }
        ksid_info_list = ksid_info_list->ksid_info_next;
    }

    // Adding new ksid_info to shared data
    qkd_ksid_info_t *new_ksid_info = malloc(sizeof(qkd_ksid_info_t));
    uuid_copy(new_ksid_info->key_stream_id, *key_stream_id);
    qkd_copy_qos(&new_ksid_info->qos, qos);
    qkd_copy_uri(&new_ksid_info->source, &source);
    qkd_copy_uri(&new_ksid_info->destination, &destination);
    new_ksid_info->ksid_info_next = NULL;

    if (ksid_info_list == NULL)
        config->ksid_info_linked_list = new_ksid_info;  // Insert as the first element of the linked list
    else
        ksid_info_list->ksid_info_next = new_ksid_info;  // Insert at the end of the linked list

    *status = qkd_status_code_successful;
}

void qkd_get_key(uuid_t key_stream_id, uint32_t *index, qkd_key_buffer_t *key_buffer, qkd_metadata_t *metadata,
                 qkd_status_t *status, void *app_context_data) {
    // Use or modify app_context_data
    server_context_data_example_t *config = app_context_data;

    char device_info[SSL3_RT_MAX_PLAIN_LENGTH];
    sprintf(device_info, "{\"device_info\": [\"key_rate\": %d, \"qber\": %f]}", config->key_rate, config->qber);
    size_t data_size = qkd_sizeof_str(device_info);
    *metadata = (qkd_metadata_t) {data_size, malloc(data_size)};
    memcpy((*metadata).data, device_info, data_size);

    qkd_ksid_info_t *ksid_info_list = config->ksid_info_linked_list;
    while (ksid_info_list != NULL) {
        if (uuid_compare(key_stream_id, ksid_info_list->key_stream_id) == 0) break;
        ksid_info_list = ksid_info_list->ksid_info_next;
    }

    if (ksid_info_list == NULL) {  // KSID not found
        *status = qkd_status_code_gk_peer_not_connected;
        return;
    }

    srand(time(NULL));
    int r = rand() % 4;
    if (r == 0) {  // Returns insufficient key error 25% of the requests
        *status = qkd_status_code_gk_insufficient_key_available;
        return;
    }

    size_t key_size = ksid_info_list->qos.key_chunk_size;
    // TODO: Add real QKD key here
    qkd_key_buffer_t key_data = gen_pseudorandom_key(ksid_info_list->source, ksid_info_list->destination, key_size,
                                                     *index);
    *key_buffer = key_data;
    *index = *index;

    *status = qkd_status_code_successful;
}

void qkd_close(uuid_t key_stream_id, qkd_status_t *status, void *app_context_data) {
    // Use or modify app_context_data
    server_context_data_example_t *config = app_context_data;
    qkd_ksid_info_t *ksid_info_list = config->ksid_info_linked_list, *ksid_info_prev = NULL;
    while (ksid_info_list != NULL) {
        if (uuid_compare(key_stream_id, ksid_info_list->key_stream_id) == 0) break;
        ksid_info_list = ksid_info_list->ksid_info_next;
        ksid_info_prev = ksid_info_list;
    }

    if (ksid_info_list == NULL) {  // KSID not found
        *status = qkd_status_code_no_qkd_connection_available;
        return;
    }

    if (ksid_info_prev == NULL)
        config->ksid_info_linked_list = ksid_info_list->ksid_info_next;  // KSID found at the beginning of the linked list
    else
        ksid_info_prev->ksid_info_next = ksid_info_list->ksid_info_next;

    if (ksid_info_list->source != NULL) free(ksid_info_list->source);
    if (ksid_info_list->destination != NULL) free(ksid_info_list->destination);
    if (ksid_info_list->qos.metadata_mimetype != NULL) free(ksid_info_list->qos.metadata_mimetype);
    free(ksid_info_list);

    *status = qkd_status_code_successful;
}

int main(int argc, char *argv[]) {
    /**
     * Note: before implementing the main remember implementing open_connect, get_key and qkd_close is Mandatory
     */

    // Certificate sending is MANDATORY for server side
    char *cert_pem = "../ssl/127.0.0.1.pem", *key_pem = "../ssl/127.0.0.1.key";

    /* Activate/deactivate client cert verification */
    char *peer_ca_pem;
//    peer_ca_pem = "../ssl/root_CA.pem";    // Client verification enabled
    peer_ca_pem = NULL;                    // Client verification disabled

    // Optional (modify server_context_data_example_t to fit your preferences or set it to NULL)
    server_context_data_example_t config = {32768, 0.49f, "database_file.db", {}, NULL};
    uuid_generate_random(config.node_uuid);

    int max_proto_version;
//    max_proto_version = TLS1_2_VERSION;
    max_proto_version = 0;

    qkd_socket_server_init("127.0.0.1", 25575, cert_pem, key_pem, peer_ca_pem, max_proto_version, &config);
    return 0;
}
